from antelope.refs.base import EntityRef
from antelope.refs import RxRef
from antelope import comp_dir
from antelope.interfaces.iforeground import ForegroundRequired
from ..fragment_flows import group_ios, ios_exchanges, FragmentInventoryDeprecated

"""
Not sure what to do about Fragment Refs, whether they belong in the main interface. I'd like to think no, but
for now we will just deprecate them and remove functionality,
"""


class ParentFragment(Exception):
    """
    a placeholder reference_entity for parent fragments
    """


class FragmentRef(EntityRef):
    """
    Fragments can lookup:
    """
    '''
    def __init__(self, *args, **kwargs):
        super(FragmentRef, self).__init__(*args, **kwargs)
        self._known_scenarios = dict()
    '''
    _etype = 'fragment'
    _ref_field = 'parent'

    def dbg_print(self, *args):
        pass

    def __init__(self, *args, flow=None, direction=None, balance_flow=None, exchange_values=None, **kwargs):
        super(FragmentRef, self).__init__(*args, **kwargs)
        self._direction = direction
        self._flow = flow
        self._is_balance = bool(balance_flow)
        self._ref_vals = dict()
        self._exch_vals = dict()
        if isinstance(exchange_values, dict):
            self._exch_vals.update(exchange_values)

        self._anchors = dict()

    @property
    def direction(self):
        return self._direction

    @property
    def is_balance(self):
        return self._is_balance

    def _query_ev(self, **kwargs):
        return self._query._perform_query('foreground', 'ev', ForegroundRequired, self, **kwargs)

    def exchange_value(self, scenario=None, observed=None):
        if scenario is None or len(scenario) == 0:
            if observed:
                scenario = '1'
            else:
                scenario = '0'
            if scenario in self._exch_vals:
                return self._exch_vals[scenario]
            else:
                self._exch_vals[scenario] = self._query_ev(observed=bool(observed))
        else:
            if scenario in self._exch_vals:
                return self._exch_vals[scenario]
            else:
                self._exch_vals[scenario] = self._query_ev(scenario=scenario)
        return self._exch_vals[scenario]

    '''
    @property
    def is_background(self):
        """
        Can't figure out whether it ever makes sense for a fragment ref to be regarded 'background'
        :return:
        """
        return T
    '''
    @property
    def reference_entity(self):
        if self._reference_entity is None:
            parent = self.get(self.reference_field)
            if parent is ParentFragment:
                self._reference_entity = parent
                return None
            elif isinstance(parent, str):
                self._reference_entity = self._query.get(parent, origin=self.origin)
            else:
                try:
                    self._reference_entity = self._query.parent(self)
                except ParentFragment:
                    self._reference_entity = ParentFragment
                    return None
        elif self._reference_entity is ParentFragment:
            return None
        return super(FragmentRef, self).reference_entity

    @property
    def flow(self):
        return self._flow

    @property
    def _addl(self):
        return 'frag'

    @property
    def name(self):
        return self['Name']

    @property
    def is_conserved_parent(self):
        return None

    def top(self):
        if self.reference_entity is None:
            return self
        return self.reference_entity.top()

    def reference(self, flow=None):
        """
        For process interoperability
        :return:
        """
        rx = RxRef(self, self.flow, comp_dir(self.direction), self.get('Comment', None), value=1.0)
        if flow is not None:
            if not rx.flow.match(flow):
                raise ValueError('%.5s: Supplied flow %s does not match fragment' % (self.uuid, flow))
        return rx

    def _load_anchors(self):
        a = self._query.anchors(self)
        for k, anchor in a.items():
            if k == 'default':
                k = None
            self._query.make_term_from_anchor(self, anchor, k)  # stores the term

    def anchors(self, **kwargs):
        """

        :param kwargs:
        :return:
        """
        for k, anchor in kwargs.items():
            if k == 'default':
                k = None
            self._query.make_term_from_anchor(self, anchor, k)  # stores the term
        if len(self._anchors) == 0:
            self._load_anchors()
        return self._anchors

    def anchor(self, scenario=None):
        if scenario in self._anchors:
            return self._anchors[scenario]
        else:
            self._load_anchors()
            return self._anchors[scenario]

    def children_with_flow(self, flow):
        return [c for c in self.child_flows if c.flow == flow]

    @property
    def child_flows(self):
        return self._query.child_flows(self)

    @property
    def is_background(self):
        return len(self.child_flows) == 0  # self._background

    def set_name(self, name, **kwargs):
        return self._query.name_fragment(self, name, **kwargs)

    '''
    Process compatibility
    '''
    def cutoffs(self, scenario=None, **kwargs):
        ios = self._query.cutoff_flows(self, scenario=scenario, **kwargs)
        return ios_exchanges(ios, ref=self)

    def activity(self, scenario=None, **kwargs):
        """
        Report interior nodes of the fragments and their activity levels-- converse of inventory()
        :return:
        """
        return self._query.activity(self, scenario=scenario, **kwargs)

    def reference_value(self, flow):
        return self._ref_vals[flow.external_ref]

    def tree(self, scenario=None, observed=False):
        return self._query.tree(self, scenario=scenario, observed=observed)

    def show_tree(self, scenario=None, observed=False):
        """
        The old show_tree finally gets properly re-implemented.
        :param scenario:
        :param observed:
        :return:
        """
        tree = self.tree(scenario=scenario, observed=observed)  # these come already sorted
        if self.reference_entity is None:
            pnts = []
        else:
            pnts = [self.reference_entity.external_ref]
        cur_stage = ''

        if observed:
            delim = '[]'
        else:
            delim = '()'

        def _pfx():
            return '    | ' * len(pnts)

        def _print_branch(_brnch, _cur_stage):
            if _brnch.group != _cur_stage:
                _cur_stage = _brnch.group
                print('   %s %5s Stage: %s' % (_pfx(), ' ', _cur_stage))
            if _brnch.magnitude is None:
                mag = '--:--'
            else:
                mag = '%7.3g' % _brnch.magnitude
            print('   %s%s%s %.5s %s %s %s%s %s' % (_pfx(), _brnch.node.dirn, _brnch.term_str,
                                                       _brnch.node.entity_uuid,
                                                       delim[0], mag, _brnch.unit, delim[1], _brnch.name))
            return _cur_stage

        for branch in tree:
            if branch.parent is None:
                # print first round
                if len(pnts) > 0:
                    raise ValueError(pnts)
                cur_stage = _print_branch(branch, cur_stage)
                pnts.append(branch.node.entity_id)
            else:
                # handle parents and print subsequent rounds
                if branch.parent != pnts[-1]:  # either up or down
                    if branch.parent in pnts:
                        while branch.parent != pnts[-1]:
                            pnts.pop()
                            print('   %s    x ' % _pfx())  # end cap
                    else:
                        print('   %s [%s]' % (_pfx(), branch.term.unit))  # new generation
                        pnts.append(branch.parent)
                cur_stage = _print_branch(branch, cur_stage)

        # finish up by capping off remaining levels
        while len(pnts) > 0:
            pnts.pop()
            print('   %s    x ' % _pfx())  # end cap

    def nodes(self, scenario=None, descend=True):
        """
        Report proximal terminal nodes for the fragment (recurse until a nondescend is reached)
        :param scenario: [None]
        :param descend: [True] if False, yield subfragments as nodes
        :return: generator of terminal nodes
        """
        term = self.anchor(scenario)
        yds = set()
        if term.is_process or term.is_context or term.is_unresolved:
            if term.term_node not in yds:
                yield term.term_node
                yds.add(term.term_node)
        elif term.is_subfrag:
            if term.descend and descend:
                for n in term.term_node.nodes(scenario, descend=descend):
                    if n not in yds:
                        yield n
                        yds.add(n)
            else:
                yield term.term_node
            # foreground, null: do nothing
        for c in self.child_flows:
            for n in c.nodes(scenario, descend=descend):
                if n not in yds:
                    yield n
                    yds.add(n)

    def traverse(self, scenario=None, **kwargs):
        return self._query.traverse(self, scenario=scenario, **kwargs)

    def lci(self, scenario=None):
        """
        TODO
        complex process that will require a recursive traversal and accumulation of LCIs from self + child fragments
        :param scenario:
        :return:
        """
        raise NotImplementedError

    def fragment_lcia(self, lcia_qty, scenario=None, mode=None, **kwargs):
        """

        :param lcia_qty:
        :param scenario:
        :param mode: None, 'detailed', 'flat', 'stage', 'anchor'
        :param kwargs:
        :return:
        """
        return self._query.fragment_lcia(self, lcia_qty, scenario=scenario, mode=mode, **kwargs)

    def bg_lcia(self, lcia_qty, scenario=None, **kwargs):
        return self.fragment_lcia(self, lcia_qty, scenario=scenario, **kwargs)

    def unit_inventory(self, scenario=None, observed=None):
        raise FragmentInventoryDeprecated('"inventory" is an exchange method. Use "unit_flows" instead.')

    def unit_flows(self, scenario=None, observed=None, frags_seen=None):
        """
        Fragment Refs do not expose internal flows during traversal (and traversal route may be protected)

        :param scenario:
        :param observed: ignored; supplied only for signature consistency
        :param frags_seen: not bothering with this for the moment (until there's a recursive crash and then what)
        :return:
        """
        '''
        return NotImplemented
        '''
        ios = self._query.cutoff_flows(self, scenario=scenario, observed=observed)
        return ios, ()

    def scenarios(self, **kwargs):
        return self._query.scenarios(self, **kwargs)



from .lc_foreground import LcForeground, FOREGROUND_ENTITY_TYPES
from .v1_client import AntelopeV1Client
from .oryx_client import OryxClient

PROVIDERS = ['AntelopeV1Client', 'LcForeground', 'OryxClient']


def init_fcn(source, ref=None, **kwargs):
    """
    Returns an LcForeground implementation
    :param source: Filename to store the serialized A and B matrices in numpy [matlab] format.
    :param ref: semantic reference for the archive
    :param kwargs:
    :return: an LcForeground archive.
    """
    if ref is None:
        ref = 'test.free.background'
    return LcForeground(source, ref=ref, **kwargs)  # make_interface('background') to generate/access flat bg




import unittest
from antelope import IndexRequired

from antelope_core import LcCatalog


cat = LcCatalog.make_tester()
ref = 'calrecycle.antelope'

cat.new_resource(ref, 'http://www.antelope-lca.net/uo-lca/api/', 'AntelopeV1Client',
                 store=False, interfaces=['basic', 'index', 'inventory', 'quantity'], quiet=True)


# @unittest.skip('"No Access to Entity" needs debugged')
class AntelopeV1Client(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.ar = cat.get_archive(ref)
    def test_stages(self):
        self.assertEqual(len(self.ar.get_endpoint('stages')), 87)

    def test_stagename(self):
        # inv = self.ar.make_interface('inventory')
        self.assertEqual(self.ar.get_stage_name('42'), 'Natural Gas')
        self.assertEqual(self.ar.get_stage_name('47'), 'Natural Gas Supply')
        self.assertEqual(self.ar.get_stage_name('81'), 'WWTP')

    def test_impactcategory(self):
        self.assertEqual(self.ar._get_impact_category(6), 'Cancer human health effects')
        with self.assertRaises(ValueError):
            self.ar._get_impact_category(5)

    def test_nonimpl(self):
        with self.assertRaises(IndexRequired):
            next(cat.query(ref).targets('flows/87'))

    def test_traversal(self):
        ffs = cat.query(ref).get('fragments/47').traverse()
        self.assertEqual(len(ffs), 14)
        self.assertSetEqual({-0.5, -0.01163, -0.0102, 0.0, 0.5}, set(round(x.node_weight, 5) for x in ffs))

    def test_lcia(self):
        lcia = cat.query(ref).get('fragments/19').fragment_lcia('lciamethods/4')
        self.assertSetEqual(set(x.external_ref for x in lcia.component_entities()),
                            {'Crude Oil', 'Electricity', 'Natural Gas', 'Refinery'})
        self.assertSetEqual(set(round(x.cumulative_result, 10) for x in lcia.components()),
                            {0.0004522897, 0.0000733389, 0.0000419222, 0.0001582613})
        self.assertAlmostEqual(lcia.total(), 0.0007258121306, places=12)


if __name__ == '__main__':
    unittest.main()



import unittest
import os
from shutil import rmtree

from ..foreground_catalog import ForegroundCatalog

WORKING_DIR = os.path.join(os.path.dirname(__file__), 'test-foreground')
test_ref = 'test.foreground'


class FgImplementationTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cat = ForegroundCatalog(WORKING_DIR)
        cls.fg = cls.cat.create_foreground(ref=test_ref)

    def test_new_flow(self):
        f = self.fg.new_flow('Test flow', 'mass')
        ar = self.cat.get_archive(self.fg.origin)
        self.assertIs(ar[f.external_ref], f)
        self.assertEqual(self.fg.get(f.external_ref), f)

    @unittest.skip('Disabled frag method for now as inside baseball')
    def test_frag(self):
        f = self.fg.new_flow('a silly flow', 'number of items')
        frag = self.fg.new_fragment(f, 'Output')
        self.assertIs(self.fg.frag(frag.uuid[:3]), frag)  ##

    def test_named_frag(self):
        f = self.fg.new_flow('a reference flow')
        frag_name = 'a stupendous fragment'
        frag = self.fg.new_fragment(f, 'Output', external_ref=frag_name)
        self.assertIs(self.fg.get(frag_name), frag)
        self.assertIs(self.fg.get(frag.uuid), frag)
        with self.assertRaises(ValueError):
            self.fg.new_fragment(f, 'Output', external_ref=frag_name)

    @classmethod
    def tearDownClass(cls):
        rmtree(WORKING_DIR)


if __name__ == '__main__':
    unittest.main()



from antelope_core.implementations import QuantityImplementation
from antelope_core.lcia_results import LciaResult

from math import isclose
from antelope import EntityNotFound

from antelope_core.exchanges import ExchangeValue
from antelope_core.characterizations import QRResult

from .exceptions import AntelopeV1Error


class FragmentFlowProxy(object):
    def __init__(self, entity_type, external_ref):
        self._etype = entity_type
        self._eref = external_ref

    @property
    def entity_type(self):
        return self._etype

    @property
    def external_ref(self):
        return self._eref

    def __str__(self):
        return '%s %s' % (self.external_ref, self.entity_type)


class AntelopeQuantityImplementation(QuantityImplementation):
    """
    Computes LCIA Results using [deserializing] antelope v1 web queries
    """
    '''
    LCIA handling
    '''
    def add_lcia_component(self, res, component):
        if 'processID' in component:
            entity = self._archive.retrieve_or_fetch_entity('processes/%s' % component['processID'])
            loc = entity['SpatialScope']
        elif 'fragmentFlowID' in component:  # not currently implemented
            entity = FragmentFlowProxy('FragmentFlow', 'fragmentflows/%s' % component['fragmentFlowID'])
            loc = 'GLO'
        elif 'fragmentStageID' in component:  # currently the default
            entity = FragmentFlowProxy('Stage', self._archive.get_stage_name(component['fragmentStageID']))
            loc = 'GLO'
        else:
            raise ValueError('Unable to handle unrecognized LCIA Result Type\n%s' % component)

        if len(component['lciaDetail']) == 0:
            res.add_summary(entity.external_ref, entity, 1.0, component['cumulativeResult'])
        else:
            if entity.entity_type != 'process':
                raise TypeError('Antelope v1 should not have details for non-process entities\n%s' % entity)
            res.add_component(entity.external_ref, entity)
            for d in component['lciaDetail']:
                self._add_lcia_detail(res, entity, d, loc=loc)

    def _add_lcia_detail(self, res, entity, detail, loc='GLO'):
        flow = self._archive.retrieve_or_fetch_entity('flows/%s' % detail['flowID'])
        cx = self._archive.tm[flow.context]
        exch = ExchangeValue(entity, flow, detail['direction'], termination=cx, value=detail['quantity'])
        qrr = QRResult(flow.flowable, flow.reference_entity, res.quantity, cx, loc, self.origin, detail['factor'])
        res.add_score(entity.external_ref, exch, qrr)

    @staticmethod
    def check_total(res, check):
        if not isclose(res, check, rel_tol=1e-8):
            raise AntelopeV1Error('Total and Check do not match! %g / %g' % (res, check))

    def get_lcia_quantity(self, quantity_ref):
        if isinstance(quantity_ref, str):
            return self._archive.retrieve_or_fetch_entity(quantity_ref)
        elif quantity_ref.origin == self._archive.ref:
            return self._archive.retrieve_or_fetch_entity(quantity_ref.external_ref)
        else:
            raise EntityNotFound

    def factors(self, quantity, flowable=None, compartment=None, **kwargs):
        """
        This needs to be done
        :param quantity:
        :param flowable:
        :param compartment:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    def profile(self, flow, **kwargs):
        f = self._archive.retrieve_or_fetch_entity(flow)
        endpoint = '%s/flowpropertymagnitudes' % flow
        fpms = self._archive.get_endpoint(endpoint)
        for cf in fpms:
            q = self._archive.retrieve_or_fetch_entity('flowproperties/%s' % cf['flowProperty']['flowPropertyID'])
            if 'location' in cf:
                location = cf['location']
            else:
                location = 'GLO'
            self.characterize(f.link, f.reference_entity, q, cf['magnitude'], context=f.context, location=location)
        for cf in self._archive.tm.factors_for_flowable(f.flowable):
            yield cf

    def lcia(self, process, ref_flow, quantity_ref, **kwargs):
        """
        Antelope v1 doesn't support or even have any knowledge of process reference-flows. this is a somewhat
        significant design flaw.  well, no matter.  each antelopev1 process must therefore represent an allocated single
        operation process that has an unambiguous reference flow.  This is a problem to solve on the server side;
        for now we just ignore the ref_flow argument.

        If the quantity ref is one of the ones natively known by the antelope server-- i.e. if it is a catalog ref whose
        origin matches the origin of the current archive-- then it is trivially used.  Otherwise, the lcia call reduces
        to obtaining the inventory and computing LCIA locally.
        :param process:
        :param ref_flow:
        :param quantity_ref:
        :param kwargs:
        :return:
        """
        lcia_q = self.get_lcia_quantity(quantity_ref)
        endpoint = '%s/%s/lciaresults' % (process, lcia_q.external_ref)
        lcia_r = self._archive.get_endpoint(endpoint, cache=False)

        res = LciaResult(lcia_q, scenario=lcia_r.pop('scenarioID'))
        total = lcia_r.pop('total')

        if len(lcia_r['lciaScore']) > 1:
            raise AntelopeV1Error('Process LCIA result contains too many components\n%s' % process)

        component = lcia_r['lciaScore'][0]
        cum = component['cumulativeResult']
        self.check_total(cum, total)

        if 'processes/%s' % component['processID'] != process:
            raise AntelopeV1Error('Reference mismatch: %s begat %s' % (process, component['processID']))

        self.add_lcia_component(res, component)

        self.check_total(res.total(), total)
        return res




from setuptools import setup, find_packages

VERSION = '0.3.3'

requires = [
    'antelope_interface>=0.2.7',
    'antelope_core>=0.3.3'
]

"""
Revision history
0.3.3   2024-09-10 - normalize how directed flows are used in lci()
                     use dependencies() instead of inventory() to extend process models
                     numerous bugfixes to observations and foreground handling

0.3.2   2024-08-15 - move fg methods to interface. lots of remote traversal work

0.3.1   2024-05-15 - "inventory" operations on fragments deprecated
                     tester catalogs now filesystem-free
                     handle unresolved anchor issues 

0.3.0   2024-01-05 - 0.3-branch development version, supporting end-user access to vault.lc resources 

# ^ 0.3.* 0.3-branch fork   
# v 0.2.* main / master for legacy projects

0.2.1 - 21 Jul 2023 - Subfrags comes home-- complete changes throughout the system, impossible to recount.

0.1.7 - 11 Aug 2021 - TRAVIS release

0.1.3 - 30 Dec 2020 - First public release
"""

setup(
    name="antelope_foreground",
    version=VERSION,
    author="Brandon Kuczenski",
    author_email="bkuczenski@ucsb.edu",
    install_requires=requires,
    url="https://github.com/AntelopeLCA/foreground",
    summary="A foreground model building implementation",
    long_description_content_type='text/markdown',
    long_description=open('README.md').read(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering"
    ],
    python_requires='>=3.6',
    packages=find_packages()
)



"""
A foreground archive is an LcArchive that also knows how to access, serialize, and deserialize fragments.
"""
import json
import os
import re
import logging

from typing import Optional

from collections import defaultdict
from pydantic import ValidationError

from ..foreground_query import DelayedQuery, ForegroundNotSafe, QueryIsDelayed, MissingResource
from ..refs.fragment_ref import FragmentRef
from ..implementations import AntelopeForegroundImplementation, AntelopeBasicImplementation
from ..models import ForegroundMetadata, ForegroundRelease, Observation

from antelope import PropertyExists, CatalogRef, EntityNotFound
from antelope_core.archives import BasicArchive, EntityExists, BASIC_ENTITY_TYPES
from ..entities.fragments import LcFragment


FOREGROUND_ENTITY_TYPES = BASIC_ENTITY_TYPES + ('fragment', )


class AmbiguousReference(Exception):
    pass


class FragmentNotFound(Exception):
    pass


class FragmentMismatch(Exception):
    pass


class NonLocalEntity(Exception):
    """
    Foregrounds should store only local entities and references to remote entities
    """
    pass


class LcForeground(BasicArchive):
    """
    An LcForeground is defined by being anchored to a physical directory, which is used to serialize the non-fragment
    entities.  Also within this directory is a subdirectory called fragments, which is used to store fragments
    individually as files.

    Fragments are "observations" of foreground systems. A flow is observed to pass from / into some reference entity
    and is terminated (other end of the flow) to x, which can either be None (a cutoff), or an entity ID which is
    dereferenced to a context (=> elementary) or a process (=>intermediate).

    A foreground is then a collection of observations of flows passing through processes.

    Foreground models can be constructed flow by flow (observed from unit process inventories
    """
    _entity_types = FOREGROUND_ENTITY_TYPES
    _ns_uuid_required = None
    _frags_loaded = False

    _observations = None

    def _ref_to_nsuuid(self, key):
        return None

    def _load_entities_json(self, filename):
        with open(filename, 'r') as fp:
            self.load_from_dict(json.load(fp), jsonfile=filename)

    def _ref_to_key(self, key):
        """
        In a foreground, there are three different ways to retrieve an entity.

        (0) by link.  The master key for each entity is its link, which is its key in _entities.

        (1) by uuid.  fg archives have no nsuuid capabilities, but if an entity has a uuid, it can be retrieved by it.

        (2) by name.  Fragments and any other entities native to the foreground archive can also be retrieved by their
        external ref.  Fragments are the only entities that can be *assigned* a name after creation.  [for this reason,
        Fragments MUST have UUIDs to be used for hashing]
        [Note: flows can also be named but their name is not their external_ref]

         * _ref_to_key transmits the user's heterogeneous input into a valid key to self._entities, or None
           - if the key is already a known link, it's easy
           - if the key is a custom name, it's almost as easy
           - if the key CONTAINS a uuid known to the db, then a link for an entity with that UUID is
              NONDETERMINISTICALLY returned
         * __getitem__ uses _ref_to_key to translate the users input into a key

        For the thorns around renaming fragments, see self.name_fragment() below

        :param key:
        :return:
        """
        if key in self._entities:
            return key
        elif key in self._ext_ref_mapping:
            return self._ext_ref_mapping[key]
        else:
            uid = self._ref_to_uuid(key)
            if uid in self._entities:
                return self._entities[uid]
        return None

    def _add_ext_ref_mapping(self, entity):
        if entity.external_ref in self._ext_ref_mapping:
            if entity == self[entity.external_ref]:
                return  # skip it
            raise EntityExists('External Ref %s already refers to %s' % (entity.external_ref, self[entity.external_ref]))
        self._ext_ref_mapping[entity.external_ref] = entity.link

    def __getitem__(self, item):
        """
        Note: this user-friendliness check adds 20% to the execution time of getitem-- so avoid it if possible
        (use _get_entity directly -- especially now that upstream is now deprecated)
        (note that _get_entity does not get contexts)

        :param item:
        :return:
        """
        if hasattr(item, 'link'):
            item = item.link
        entity = super(BasicArchive, self).__getitem__(item)
        if entity:
            if entity.link in self._unresolved and self._catalog is not None:
                return self._replace_unresolved(entity)
        return entity

    def _replace_unresolved(self, entity):
        # try to find a new one:
        new = self.catalog_ref(entity.origin, entity.external_ref, entity_type=entity.entity_type)
        if not new.resolved:
            return entity  # nothing to do
        ks = [k for k, v in self._entities.items() if v is entity]
        self._unresolved.remove(entity.link)
        for k in ks:
            self._entities[k] = new

        if entity.link != new.link:
            self._ext_ref_mapping[entity.external_ref] = new.link
        return new

    @property
    def _archive_file(self):
        return os.path.join(self.source, 'entities.json')

    @property
    def _metadata_file(self):
        return os.path.join(self.source, 'metadata.json')

    @property
    def _fragment_dir(self):
        return os.path.join(self.source, 'fragments')

    def __init__(self, fg_path, catalog=None, **kwargs):
        """

        :param fg_path:
        :param catalog: A foreground archive requires a catalog to deserialize saved fragments. If None, archive will
        still initialize (and will even be able to save fragments) but non-locally terminated fragments will fail.
        :param ns_uuid: Foreground archives may not use ns_uuids, so any namespace uuid provided will be ignored.
        :param kwargs:
        """
        if catalog is not None:
            kwargs['term_manager'] = kwargs.pop('term_manager', catalog.lcia_engine)
        super(LcForeground, self).__init__(fg_path, **kwargs)
        self._catalog = catalog
        self._ext_ref_mapping = dict()
        self._frags_with_flow = defaultdict(set)

        self._observations = []

        self._metadata = None

        # self._delayed_refs = []
        self._unresolved = set()

        self.load_all()

    def observations(self, fresh=True):
        if fresh:
            for obs in self._observations:
                yield obs
        else:
            for frag in self.fragments():
                for obs in frag.observations():
                    yield obs

    @property
    def metadata(self):
        return self._metadata

    def catalog_ref(self, origin, external_ref, entity_type=None, **kwargs):
        """

        :param origin:
        :param external_ref:
        :param entity_type:
        :param kwargs:
        :return:
        """
        '''#this is totally immaterial
        if entity_type == 'term':
            if origin in self._catalog.foregrounds:
                entity_type = 'fragment'
            else:
                entity_type = 'process'
        '''
        if origin in self.catalog_names:
            return self.get(external_ref)
        try:
            return self._catalog.internal_ref(self.ref, origin, external_ref)
        except (ForegroundNotSafe, MissingResource):
            print('{%s} Creating delayed ref %s/%s [%s]' % (self.ref, origin, external_ref, entity_type))
            dq = DelayedQuery(origin, self._catalog, self.ref)
            if entity_type == 'fragment':
                return FragmentRef(external_ref, dq, **kwargs)
            self._counter['delayed'] += 1
            return CatalogRef.from_query(external_ref, query=dq, etype=entity_type, **kwargs)
        except EntityNotFound:
            print('{%s} entity %s/%s not found' % (self.ref, origin, external_ref))
            return CatalogRef(origin, external_ref, entity_type=entity_type, **kwargs)

    @property
    def delayed(self):
        return self._counter['delayed']

    @property
    def unresolved(self):
        return list(self._unresolved)

    def catalog_query(self, origin, **kwargs):
        return self._catalog.query(origin, **kwargs)

    def _fetch(self, entity, **kwargs):
        return self.__getitem__(entity)

    def _load_all(self):
        if os.path.exists(self._archive_file):
            self._load_entities_json(self._archive_file)
        if os.path.exists(self._metadata_file):
            try:
                with open(self._metadata_file) as fp:
                    self._metadata = ForegroundMetadata(**json.load(fp))
            except (ValidationError, json.JSONDecodeError):
                self._metadata = ForegroundMetadata(version_major=-1, version_minor=-1,
                                                    dataSource=self.source,
                                                    description='failed to load',
                                                    author=self._metadata_file)
        else:
            self._metadata = ForegroundMetadata(version_major=0, version_minor=0,
                                                dataSource=self.source,
                                                description='LcForeground',
                                                author='Antelope')

    def make_interface(self, iface):

        if iface == 'foreground' or iface == 'basic':
            if self._frags_loaded is False:
                self._load_fragments()
                self.check_counter('fragment')
                self._frags_loaded = True
            if iface == 'foreground':
                return AntelopeForegroundImplementation(self)
            else:
                return AntelopeBasicImplementation(self)
        else:
            return super(LcForeground, self).make_interface(iface)

    def set_origin(self, origin):
        super(LcForeground, self).set_origin(origin)
        for k in self.entities_by_type('fragment'):
            k._origin = origin

    def _flow_ref_from_json(self, e, external_ref):
        origin = e.pop('origin')
        r_q = self[e.pop('referenceQuantity', None)]  # quantity must have been loaded
        ref = self.catalog_ref(origin, external_ref, entity_type='flow', reference_entity=r_q, **e)
        if not ref.resolved and self._frags_loaded:  # not found
            try:
                ref_q = self._catalog.get_canonical(r_q)
                name = e.pop('Name', None) or 'unnamed flow %s' % origin
                ref = self.make_interface('foreground').add_or_retrieve(external_ref, ref_q, name, **e)
            except EntityNotFound:
                pass
        return ref

    def _make_entity(self, e, etype, ext_ref):
        if e['origin'] != self.ref and e['origin'] not in self.catalog_names:
            if etype == 'flow':
                return self._flow_ref_from_json(e, ext_ref)
            elif etype == 'quantity':
                unit = e.pop('referenceUnit', None)
                return self.catalog_ref(e.pop('origin'), ext_ref, entity_type='quantity', reference_entity=unit, **e)
        e.pop('origin', None)  # just go ahead and domesticate anything we make as an entity
        return super(LcForeground, self)._make_entity(e, etype, ext_ref)

    def add(self, entity):
        """
        Reimplement base add to (1) use link instead of external_ref, (2) merge instead of raising a key error.
        not sure we really want/need to do the merge thing- we will find out
        :param entity:
        :return:
        """
        if entity.origin is None:
            entity.origin = self.ref  # have to do this now in order to have the link properly defined
        elif entity.is_entity:
            if entity.origin != self.ref:
                if entity.origin not in self.catalog_names:
                    # TODO: Alert! entity properties are not preserved in the local ref
                    entity = self.catalog_ref(entity.origin, entity.external_ref, entity_type=entity.entity_type)
            elif entity.entity_type == 'quantity':
                q_ref = entity.make_ref(self.query)
                self._catalog.register_entity_ref(q_ref)

            # for p in entity.properties:
            #     enew[p] = entity[p]  ...
        try:
            self._add(entity, entity.link)
            if not entity.is_entity and not entity.resolved:
                self._unresolved.add(entity.link)
        except EntityExists:
            if entity is self[entity.link]:
                pass
            elif entity.entity_type == 'fragment' and entity.external_ref != entity.uuid:
                raise ValueError('Name is already taken: %s' % entity.external_ref)
            else:
                # merge incoming entity's properties with existing entity
                current = self[entity.link]
                current.merge(entity)

        if hasattr(entity, 'uuid') and entity.uuid is not None:
            self._entities[entity.uuid] = entity

        if entity.origin == self.ref and entity.external_ref != entity.uuid:
            self._add_ext_ref_mapping(entity)

        # it's up to the other foregrounds (local or not) to add their own quantities to the TM
        if entity.entity_type == 'quantity':
            if entity.origin != self.ref and entity.origin in self._catalog.foregrounds:
                # do not add to tm
                return

        try:
            self._add_to_tm(entity)  # , merge_strategy='distinct')  # DWR!!! need to
        except QueryIsDelayed:
            pass

        if entity.entity_type == 'fragment':
            self._frags_with_flow[entity.flow].add(entity)

    def _add_children(self, entity):
        if entity.entity_type == 'fragment':
            self.add_entity_and_children(entity.flow)
            for c in entity.child_flows:
                self.add_entity_and_children(c)
        else:
            super(LcForeground, self)._add_children(entity)

    def _rename_mechanics(self, frag, oldname):
        """
        This function updates all the various mappings between a fragment and its references,
        EXCEPT the ext_ref_mapping which is left to outside code because of complexities with sequencing

        This can surely be simplified (currently, renaming a fragment requires calling this twice) but not today.

        :param frag:
        :param oldname:
        :return:
        """
        self._entities[frag.link] = self._entities.pop(oldname)
        self._ents_by_type['fragment'].remove(oldname)
        self._ents_by_type['fragment'].add(frag.link)

    def _dename_fragment(self, prior):
        name = prior.external_ref
        if name == prior.uuid:
            return  # nothing to do
        print('removing name from fragment %s to %s' % (name, prior.uuid))
        priorlink = prior.link
        prior.de_name()
        self._ext_ref_mapping.pop(name)

        self._rename_mechanics(prior, priorlink)
        assert self._ref_to_key(name) is None
        obs = Observation.dename(prior)
        self._observations.append(obs)
        return obs

    def name_fragment(self, frag, name, auto=None, force=None):
        """
        This function is complicated because we have so many dicts:
         _entities maps link to entity
         _ext_ref_mapping maps custom name to link
         _ents_by_type keeps a set of links by entity type

        So, when we name a fragment, we want to do the following:
         - ensure the fragment is properly in the _entities dict
         - ensure the name is not already taken
         - set the fragment entity's name
         * pop the old link and replace its object with the new link in _entities
         * remove the old link and replace it with the new link in ents_by_type['fragment']
         * add the name with the new link to the ext_ref_mapping
        :param frag:
        :param name:
        :param auto: if True, if name is taken, apply an auto-incrementing numeric suffix until a free name is found
        :param force: if True, if name is taken, de-name the prior fragment and assign the name to the current one.
         This requires:
           - first de-naming the prior fragment,
           - then swapping its _entities entry
           - then swapping its _ents_by_type entry
           - then removing its duplicate ext_ref_mapping
           = then proceeding to normally rename the new frag
         If both auto and force are specified, auto takes precedence.
        :return: returns the assigned name
        """
        if frag.external_ref == name:
            return  # nothing to do-- fragment is already assigned that name
        if name.find('/') >= 0:
            raise ValueError('"%s": Fragment name cannot include forward slash' % name)
        current = self[frag.link]
        if current is not frag:
            if current is None:
                raise FragmentNotFound(frag)
            raise FragmentMismatch('%s\n%s' % (current, frag))
        if self._ref_to_key(name) is not None:
            if auto:
                inc = 0
                newname = '%s %02d' % (name, inc)
                while self._ref_to_key(newname) is not None:
                    inc += 1
                    newname = '%s %02d' % (name, inc)
                name = newname
            elif force:
                prior = self._entities[self._ref_to_key(name)]
                assert prior.external_ref == name
                self._dename_fragment(prior)
            else:
                raise ValueError('Name is already taken: "%s"' % name)
        oldname = frag.link
        try:
            frag.external_ref = name  # will raise PropertyExists if already set
        except PropertyExists:  # don't need force to rename a nonconflicting fragment
            self._dename_fragment(frag)
            oldname = frag.link
            frag.external_ref = name

        self._add_ext_ref_mapping(frag)
        self._rename_mechanics(frag, oldname)
        assert self._ref_to_key(name) == frag.link
        obs = Observation.from_name(frag)
        self._observations.append(obs)

        return obs

    def observe_ev(self, fragment, scenario=None, value=None, units=None):
        if value is not None:
            fragment.observe(scenario=scenario, value=value, units=units)
            obs = Observation.ev(fragment, scenario, value, units)
            self._observations.append(obs)
        elif scenario is None and fragment.observed_ev == 0:
            # If we are observing a fragment with no scenario and no observed ev, we simply apply the cached ev
            fragment.observe(value=fragment.cached_ev)
            obs = Observation.ev(fragment, scenario, fragment.cached_ev)
            self._observations.append(obs)
        else:
            fragment.set_exchange_value(scenario, None)
            obs = Observation.ev(fragment, scenario, None)
        return obs

    def observe_anchor(self, fragment, scenario, anchor_target, anchor_flow, descend=None):
        if anchor_target.entity_type != 'context':
            # adds dependency and turns models into entities (idempotent)
            anchor_target = self._catalog.internal_ref(self.ref, anchor_target.origin, anchor_target.external_ref)
        term = fragment.terminate(anchor_target, scenario=scenario, term_flow=anchor_flow, descend=descend)
        anchor = term.to_anchor()
        obs = Observation.from_anchor(fragment, scenario, anchor)  # and back into model
        self._observations.append(obs)
        return obs

    '''
    Save and load the archive
    '''
    def _do_load(self, fragments):
        for f in fragments:
            frag = LcFragment.from_json(self, f)
            # if frag.external_ref != frag.uuid:  # don't need this anymore- auto-added for entities with self.ref
            #     self._add_ext_ref_mapping(frag)
            self.add(frag)

        self._frags_loaded = True  # they all now exist and should not be loaded again

        for f in fragments:
            frag = self[f['entityId']]
            try:
                frag.finish_json_load(self, f)
            except AttributeError:
                print(f)
                raise

    def _load_fragments(self):
        """
        This must be done in two steps, since fragments refer to other fragments in their definition.
        First step: create all fragments.
        Second step: set reference entities and terminations
        :return:
        """
        fragments = []
        if not os.path.exists(self._fragment_dir):
            if self._catalog and self._catalog.test:
                return
            os.makedirs(self._fragment_dir)
        for file in os.listdir(self._fragment_dir):
            if os.path.isdir(os.path.join(self._fragment_dir, file)):
                continue
            with open(os.path.join(self._fragment_dir, file), 'r') as fp:
                j = json.load(fp)

            fragments.extend(j['fragments'])
        self._do_load(fragments)

    def _recurse_frags(self, frag):
        frags = [frag]
        for x in sorted(frag.child_flows, key=lambda z: z.uuid):
            frags.extend(self._recurse_frags(x))
        return frags

    def save_fragments(self, save_unit_scores=False):
        current_files = os.listdir(self._fragment_dir)
        for r in self._fragments():
            frags = [t.serialize(save_unit_scores=save_unit_scores) for t in self._recurse_frags(r)]
            fname = r.uuid + '.json'
            if fname in current_files:
                current_files.remove(fname)
            tgt_file = os.path.join(self._fragment_dir, fname)
            with open(tgt_file, 'w') as fp:
                json.dump({'fragments': frags}, fp, indent=2, sort_keys=True)
        for leftover in current_files:
            if not os.path.isdir(os.path.join(self._fragment_dir, leftover)):
                print('deleting %s' % leftover)
                os.remove(os.path.join(self._fragment_dir, leftover))

    def save_metadata(self):
        if not os.path.isdir(self.source):
            if self._catalog and self._catalog.test:
                return
            os.makedirs(self.source)

        with open(self._metadata_file, 'w') as fp:
            json.dump(self._metadata.model_dump(), fp, indent=2)

    def save(self, save_unit_scores=False):
        if not os.path.isdir(self.source):
            if self._catalog and self._catalog.test:
                logging.error('Cannot save new foregrounds during tester operation')
                return False
            os.makedirs(self.source)

        self.write_to_file(self._archive_file, gzip=False, characterizations=True, values=True, domesticate=False)

        self.save_metadata()

        if not os.path.isdir(self._fragment_dir):
            os.makedirs(self._fragment_dir)
        self.save_fragments(save_unit_scores=save_unit_scores)
        return True

    def update_metadata(self, release: Optional[ForegroundRelease] = None, bump_version=True):
        """
        update foreground metadata with release info (default is to bump the version number). also writes the
        metadata file.
        Here we assume the ref has NOT been decorated with the semantic version. add logic to prevent this if
        bad shit starts happening
        :param bump_version: [True] whether to bump the revision
        :param release: ForegroundRelease applied to the metadata
        :return: the path to the zipfile
        """

        # update metadata
        if release is None:
            release = ForegroundRelease(description='no info')

        if bump_version:
            if release.major:
                self.metadata.version_major += 1
                self.metadata.version_minor = 0
            else:
                self.metadata.version_minor += 1

        # we do in fact want to update these fields even if they are None
        self.metadata.release_notes = release.notes
        self.metadata.author = release.author
        self.metadata.description = release.description

        self.save_metadata()

    def clear_unit_scores(self, lcia_method=None):
        logging.info('Clearing unit scores in %s for %s' % (self.ref, lcia_method))
        for f in self.entities_by_type('fragment'):
            for s, t in f.terminations():
                if lcia_method is None:
                    t.clear_score_cache()
                else:
                    t.reset_score(lcia_method)
    '''
    Retrieve + display fragments
    '''
    def _fragments(self, background=None, **kwargs):
        for f in self.search('fragment', **kwargs):
            if f.reference_entity is None:
                if background is None or f.is_background == background:
                    yield f

    def _show_frag_children(self, frag, level=0, show=False):
        level += 1
        for k in frag.child_flows:
            if show:
                print('%s%s' % ('  ' * level, k))
            else:
                yield k
            for j in self._show_frag_children(k, level, show=show):
                yield j

    def fragments(self, *args, show_all=False, background=None, show=False, **kwargs):
        """
        :param : optional first param is filter string-- note: filters only on reference fragments!
        :param show_all: show child fragments as well as reference fragments
        :param background: [None] if True or False, show fragments whose background status is as specified
        :param show: [False] if true, print the fragments instead of returning them
        :param kwargs: search parameters
        :return:
        """
        for f in sorted([x for x in self._fragments(background=background, **kwargs)], key=lambda x: x.is_background):
            if len(args) != 0:
                if not bool(re.search(args[0], str(f), flags=re.IGNORECASE)):
                    continue
            if show:
                print('%s' % f)
                if show_all:
                    self._show_frag_children(f, show=show)
            else:
                yield f
                if show_all:
                    for k in self._show_frag_children(f):
                        yield k

    def fragments_with_flow(self, flow):
        for k in self._frags_with_flow[flow]:
            yield k

    def frag(self, string, many=False, strict=True):
        """
        strict=True is slow
        Works as an iterator. If nothing is found, raises StopIteration. If multiple hits are found and strict is set,
        raises Ambiguous Reference.
        :param string:
        :param many: [False] if true, generate
        :param strict: [True] synonym for not many
        :return:
        """
        if strict and not many:
            k = [f for f in self.fragments(show_all=True) if f.uuid.startswith(string.lower())]
            if len(k) > 1:
                for i in k:
                    print('%s' % i)
                raise AmbiguousReference(string)
            try:
                return k[0]
            except IndexError:
                raise FragmentNotFound(string)
        else:
            return self.frags(string)

    def frags(self, string):
        for f in self.fragments(show_all=True):
            if f.uuid.startswith(string.lower()):
                yield f

    def draw(self, string, **kwargs):
        if not isinstance(string, LcFragment):
            string = self.frag(string)
        string.show_tree(**kwargs)

    '''
    Utilities for finding terminated fragments and deleting fragments
    '''
    def delete_fragment(self, frag):
        """
        Need to purge the fragment from:
         _entities
         _ents_by_type
         _ext_ref_mapping
         _frags_with_flow
        and correct _counter.
        The fragment is not destroyed- and can be re-added. Deleting child fragments is left to interface code (why?)
        :param frag:
        :return:
        """
        if frag.origin != self.ref:
            raise FragmentMismatch('Fragment belongs to another foreground!')
        if frag.reference_entity is not None:
            frag.unset_parent()
        self._entities.pop(frag.link)
        self._ents_by_type['fragment'].remove(frag.link)
        self._counter['fragment'] -= 1
        if self._entities[frag.uuid] is frag:
            self._entities.pop(frag.uuid)
        self._ext_ref_mapping.pop(frag.external_ref, None)
        self._frags_with_flow[frag.flow].remove(frag)

    def _del_f(self, f):
        print('Deleting %s' % f)
        del self._entities[f.uuid]

    def del_orphans(self, for_real=False):
        """
        self is a foreground archive -- delete
        """
        for f in self._fragments(background=True):
            if f.reference_entity is not None:
                continue
            try:
                next(self._find_links(f))
                print('Found a link for %s' % f)
            except StopIteration:
                print('### Found orphan %s' % f)
                if for_real:
                    self._del_f(f)

    def _find_links(self, frag):
        for i in self.fragments(show_all=True):
            for s, t in i.terminations():
                if t.term_node is frag:
                    yield t

    def linked_terms(self, frag):
        """
        returns a list of terminations that match the input.
        :param frag:
        :return:
        """
        return [f for f in self._find_links(frag)]



"""
I don't really want a foreground interface--- I want a foreground construction yard.  the query is not the right
route for this maybe?

I want to do the things in the editor: create and curate flows, terminate flows in processes, recurse on processes.

anything else?
child fragment flow generators and handlers. scenario tools.

scenarios currently live in the fragments- those need to be tracked somewhere. notably: on traversal, when parameters
are encountered.

Should that be here? how about monte carlo? where does that get done?

when do I go to bed?
"""

from antelope import ForegroundInterface


class ForegroundRequired(Exception):
    pass


_interface = 'foreground'


class AntelopeForegroundInterface(ForegroundInterface):
    """
    The bare minimum foreground interface allows a foreground to return terminations and to save anything it creates.
    erm "anchors"
    """
    '''
    Left to subclasses
    '''
    def fragments(self, show_all=False, **kwargs):
        if show_all:
            raise ValueError('Cannot retrieve non-parent fragments via interface')
        for i in self._perform_query(_interface, 'fragments', ForegroundRequired,
                                     **kwargs):
            yield self.make_ref(i)

    def top(self, fragment, **kwargs):
        """
        Return the reference fragment that is top parent of named fragment
        :param fragment:
        :param kwargs:
        :return:
        """
        return self.make_ref(self._perform_query(_interface, 'top', ForegroundRequired, fragment, **kwargs))

    def parent(self, fragment, **kwargs):
        """
        A foreground-specific hook for get_reference
        :param fragment:
        :param kwargs:
        :return:
        """
        return self.make_ref(self._perform_query(_interface, 'get_reference', ForegroundRequired, fragment, **kwargs))

    def frag(self, string, many=False, **kwargs):
        """
        Return the unique fragment whose ID starts with string.

        Default: If the string is insufficiently specific (i.e. there are multiple matches), raise
        :param string:
        :param many: [False] if true, return a generator and don't raise an error
        :param kwargs:
        :return:
        """
        return self._perform_query(_interface, 'frag', ForegroundRequired,
                                   string, many=many, **kwargs)

    def frags(self, string, **kwargs):
        """
        Among only named fragments, return fragments whose names (external_ref) begin with string
        :param string:
        """
        return self._perform_query(_interface, 'frags', ForegroundRequired,
                                   string, **kwargs)

    '''
    def name_fragment(self, fragment, name, auto=None, force=None, **kwargs):
        """
        Assign a fragment a non-UUID external ref to facilitate its easy retrieval.  I suspect this should be
        constrained to reference fragments.  By default, if the requested name is taken, a ValueError is raised
        :param fragment:
        :param auto: if True, if name is taken, apply an auto-incrementing numeric suffix until a free name is found
        :param force: if True, if name is taken, de-name the prior fragment and assign the name to the current one
        :return:
        """
        return self._perform_query(_interface, 'name_fragment', ForegroundRequired,
                                   fragment, name, **kwargs)
    '''

    '''
    def find_or_create_term(self, exchange, background=None):
        """
        Finds a fragment that terminates the given exchange
        :param exchange:
        :param background: [None] - any frag; [True] - background frag; [False] - foreground frag
        :return:
        """
        return self._perform_query(_interface, 'find_or_create_term', ForegroundRequired,
                                   exchange, background=background)
    '''
    def add_or_retrieve(self, external_ref, reference, name, group=None, **kwargs):
        """
        Gets an entity with the given external_ref if it exists, , and creates it if it doesn't exist from the args.

        Note that the full spec is mandatory so it could be done by tuple.
        :param external_ref:
        :param reference:
        :param name:
        :param group: None
        :param kwargs:
        :return:
        """
        return self.make_ref(self._perform_query(_interface, 'add_or_retrieve', ForegroundRequired,
                                                 external_ref, reference, name, group=group, **kwargs))

    def post_entity_refs(self, entity_refs, **kwargs):
        """
        Add entities to the foreground by reference. mostly useful when working with remote foregrounds.
        :param entity_refs:
        :param kwargs:
        :return:
        """
        return self._perform_query(_interface, 'post_entity_refs', ForegroundRequired,
                                   entity_refs, **kwargs)

    def create_fragment_from_node(self, process_ref, ref_flow=None, include_elementary=False):
        """
        a synonym for create_process_model
        :param process_ref: a ProcessRef
        :param ref_flow:
        :param include_elementary:
        :return:
        """
        return self._perform_query(_interface, 'create_process_model', ForegroundRequired,
                                   process_ref, ref_flow=ref_flow, include_elementary=include_elementary)

    def clone_fragment(self, frag, tag=None, **kwargs):
        """

        :param frag: the fragment (and subfragments) to clone
        :param kwargs: tag - appended to all named child fragments
        :return:
        """
        return self._perform_query(_interface, 'clone_fragment', ForegroundRequired,
                                   frag, tag=tag, **kwargs)

    def delete_fragment(self, fragment, **kwargs):
        """
        Remove the fragment and all its subfragments from the archive (they remain in memory)
        This does absolutely no safety checking.

        :param fragment:
        :param kwargs:
        :return:
        """
        return self._perform_query(_interface, 'delete_fragment', ForegroundRequired,
                                   fragment, **kwargs)

    def scenarios(self, fragment, recurse=True, **kwargs):
        """
        Return a recursive list
        :param fragment:
        :param recurse: [True] whether to include scenarios in child fragments
        :param kwargs:
        :return:
        """
        return self._perform_query(_interface, 'scenarios', ForegroundRequired,
                                   fragment, recurse=recurse, **kwargs)

    def nodes(self, origin=None, scenario=None, **kwargs):
        """
        Return non-trivial anchors in the current foreground, optionally filtered by origin
        :param origin: [None] if present, only return nodes whose anchor has the given origin
        :param scenario: [None] if present, only return nodes for the given scenario
        :return:
        """
        return self._perform_query(_interface, 'nodes', ForegroundRequired, origin=origin, scenario=scenario, **kwargs)

    def knobs(self, search=None, **kwargs):
        """
        Return a list of named fragments whose values can be observed to define scenarios.  Generates a list
        of non-reference fragments with names
        :return:
        """
        return self._perform_query(_interface, 'knobs', ForegroundRequired,
                                   search=search, **kwargs)

    def set_balance_flow(self, fragment, **kwargs):
        """
        This should get folded into observe
        Specify that a given fragment is a balancing flow for the parent node, with respect to the specified fragment's
        flow's reference quantity.

        :param fragment:
        :param kwargs:
        :return:
        """
        return self._perform_query(_interface, 'set_balance_flow', ForegroundRequired,
                                   fragment, **kwargs)

    def unset_balance_flow(self, fragment, **kwargs):
        """
        This should get folded into observe
        Specify that a given fragment's balance status should be removed.  The fragment's observed EV will remain at
        the most recently observed level.
        :param fragment:
        :param kwargs:
        :return:
        """
        return self._perform_query(_interface, 'unset_balance_flow', ForegroundRequired,
                                   fragment, **kwargs)

    def create_process_model(self, process, ref_flow=None, set_background=None, **kwargs):
        """
        Create a fragment from a process_ref.  If process has only one reference exchange, it will be used automatically.
        By default, a child fragment is created for each exchange not terminated to context, and exchanges terminated
        to nodes are so terminated in the fragment.
        :param process:
        :param ref_flow: specify which exchange to use as a reference
        :param set_background: [None] Deprecated. All terminations are "to background".
        :param kwargs:
        :return:
        """
        return self._perform_query(_interface, 'create_process_model', ForegroundRequired,
                                   process, ref_flow=ref_flow, **kwargs)

    def extend_process(self, fragment, scenario=None, include_context=False, **kwargs):
        """

        :param fragment:
        :param scenario:
        :param include_context:
        :param kwargs:
        :return:
        """
        return self._perform_query(_interface, 'extend_process', ForegroundRequired,
                                   fragment, scenario=scenario, include_context=include_context, **kwargs)

    def fragment_from_exchanges(self, exchanges, parent=None, include_context=False, **kwargs):
        """

        :param exchanges:
        :param parent: if parent is None, the first exchange is taken to be a reference fragment
        :param include_context: [False] if true, create subfragments terminating to context for elementary flows.
         otherwise leaves them unspecified (fragment LCIA includes unobserved exchanges)
        :param kwargs:
        :return:
        """
        return self._perform_query(_interface, 'fragment_from_exchanges', ForegroundRequired,
                                   exchanges, parent=parent, include_context=include_context, **kwargs)



"""
This module includes both CalRecycle data source unittests and also integration tests that use the CalRecycle
data source as a fixture.
"""

import unittest
import time

from math import floor

from antelope_core import LcCatalog

from antelope_core.data_sources.local import RESOURCES_CONFIG, check_enabled
from antelope_core.catalog.catalog_root import CATALOG_ROOT

from ..calrecycle_lca import CalRecycleConfig
from requests import ConnectionError as requestsConnectionError

_debug = False

if __name__ == '__main__':
    _run_calrecycle = check_enabled('calrecycle')
else:
    _run_calrecycle = check_enabled('calrecycle') or _debug

official_start_time = time.time()

if _run_calrecycle:
    cat = LcCatalog(CATALOG_ROOT, strict_clookup=False)
    crc = CalRecycleConfig(RESOURCES_CONFIG['calrecycle']['data_root'])
    fg = crc.foreground(cat, fg_path='uolca')
    if fg.frag('aed7a').is_background:
        fg.frag('aed7a')._background = False  # postfix defect in repo generation
    if 'calrecycle.antelope' not in cat.origins:
        cat.new_resource('calrecycle.antelope', 'http://www.antelope-lca.net/uo-lca/api/', 'AntelopeV1Client',
                         store=False, interfaces=['index', 'inventory', 'quantity'], quiet=True)
    cr_time = time.time()
    print('CalRecycle loaded: %.3f' % (cr_time - official_start_time))
    try:
        ar = cat.get_archive('calrecycle.antelope')
        print('Antelope archive loaded: %.3f' % (time.time() - cr_time))
    except requestsConnectionError:
        print('HTTP connection error: Bypassing remote validation')
        ar = None

else:
    fg = None


def _get_antelope_result(frag_id, lcia_id):
    key = 'fragments/%d' % frag_id
    lcia = 'lciamethods/%d' % lcia_id
    return cat.query('calrecycle.antelope').get(key).fragment_lcia(lcia)


@unittest.skipUnless(_run_calrecycle, "CalRecycle test not enabled")
class CalRecycleTest(unittest.TestCase):
    def setUp(self):
        self.startTime = time.time()

    @staticmethod
    def _time(_id, starttime):
        t = time.time() - starttime
        print("------ %+60.60s: %.3f\n" % (_id, t))

    def tearDown(self):
        self._time('.'.join(self.id().split('.')[-2:]), self.startTime)

    @classmethod
    def setUpClass(cls):
        cls._time('starting off', official_start_time)
        cls.gwp = cat.query('calrecycle.lcia.elcd').get('370960f4-0a3a-415d-bf3e-e5ce63160bb9')
        cls.mep = cat.query('calrecycle.lcia.elcd').get('5296e2be-060b-4e50-b033-d45f85f6ac92')
        cls.criteria = cat.query('calrecycle.lcia.traci').get('de6753e3-2e0a-4fb8-b113-331735e26e3d')
        cls.lcia_map = {2: cls.gwp, 4: cls.mep, 16: cls.criteria}


    def test_create_and_load(self):
        self.assertEqual(fg.count_by_type('fragment'), 888)

    def test_scenarios(self):
        uom = fg['fragments/55']
        self.assertSetEqual(set(uom.scenarios()), {'sp24', 'quell_eg'})

    def test_named_fragment(self):
        self.assertEqual(fg['fragments/43']['Name'], 'Fuels Displacement Mixer - distillate')

    def test_passthrough(self):
        lm = fg['fragments/3']
        inv, _ = lm.unit_flows(None, observed=True)
        self.assertEqual(len(inv), 3)  # 3 inventory flows
        self.assertEqual(len(set(x.fragment.flow for x in inv)), 2)  # 2 distinct flows
        self.assertSetEqual(set(x.fragment.direction for x in inv if x.fragment.flow == lm.flow),
                            {'Input', 'Output'})  # 2 distinct directions for reference flow
        self.assertSetEqual(set(x.magnitude for x in inv), {1})  # all flows have magnitude 1

    def test_traversal(self):
        uom = fg['fragments/55']
        ffs = uom.traverse(None, observed=True)
        self.assertEqual(len(ffs), 322)  # some ffs went away with uslci bg change 43d35c7; a few came back with 110321a
        inv = uom.cutoffs(None, observed=True)
        light_fuel = next(x for x in inv if x.flow['Name'].startswith('Light Fuel'))
        self.assertEqual(floor(light_fuel.value), 116915718)
        incin = next(x for x in inv if x.flow['Name'].startswith('UO to Incin'))
        self.assertEqual(floor(incin.value), 484433)

    def test_lcia_0_indicators(self):
        """
        Investigation into gwp flowables from pre- to post-context-refactor:
        pre: 100 flowables,
         * includes two distinct 'nitrous oxide' entries with 10024-97-2 and 10024-97-3
         * includes 'carbon dioxide (biogenic)' as well as 'carbon dioxide (fossil)' and 'carbon dioxide'
         * includes 'dichloromethane' and 'cfc-12'
        post: 98 flowables
         * excludes 'dichlorodifluoromethane' and 'methylenechloride'
         * only has 'carbon dioxide' and 'carbon dioxide (fossil)
         * only has one 'nitrous oxide' (as correct)
        assessment:
         = 'dichlorodifluoromethane' and 'cfc-12' are the same
         = 'methylenechloride' and 'dichloromethane' are the same

        So it's an open question whether the biogenic CO2 situation is a problem or not since it's not yet tested.
        but it's explicable.
        :return:
        """
        self.assertEqual(len([x for x in self.gwp.flowables()]), 98)  # was 100- what's missing?
        # self.assertEqual(len([x for x in self.gwp.flowables()]), 100)
        self.assertEqual(len([x for x in self.criteria.flowables()]), 6)

    def test_lcia_1_rere(self):
        re_re = fg['fragments/52'].term
        res = re_re.term_node.fg_lcia(self.gwp, ref_flow=re_re.term_flow)
        self.assertAlmostEqual(res.total(), 0.32561424, places=6)
        res = re_re.term_node.fg_lcia(self.criteria, ref_flow=re_re.term_flow)
        self.assertAlmostEqual(res.total(), 0.00043044969, places=9)

    def test_lcia_2_frag_elec(self):
        elec = fg['fragments/1']
        res = elec.fragment_lcia(self.gwp)
        self.assertAlmostEqual(res.total(), 0.165043782, places=8)
        res = elec.fragment_lcia(self.criteria)
        self.assertAlmostEqual(res.total(), 0.0003245430, places=9)

    def test_lcia_2_frag_uom(self):
        uom = fg['fragments/55']
        res = uom.fragment_lcia(self.gwp)
        # self.assertEqual(floor(res.total()), -94202323)  # legacy Antelope result-- eg fail; private processes differ
        self.assertEqual(floor(res.total()), -94671945)  # current correct value 2018/12/28
        res_eg = uom.fragment_lcia(self.gwp, scenario='quell_eg')
        self.assertEqual(floor(res_eg.total()), -94357314)  # quell_eg value 2018/12/28
        res = uom.fragment_lcia(self.criteria, scenario='quell_eg')
        self.assertEqual(floor(res.total()), 25278)  # current quell_eg value 2018/12/28

    def test_antelope_frag_2_ng_conserv(self):
        if ar is None:
            self.skipTest('antelope archive not available')
        ng2 = fg['fragments/2']
        ffs_loc = [f for f in ng2.traverse(observed=True) if f.is_conserved]
        ffs_rem = [f for f in cat.query('calrecycle.antelope').get('fragments/2').traverse() if f.is_conserved]
        self.assertEqual(len(ffs_loc), len(ffs_rem))
        for ff in ffs_rem:
            ffl = next(f for f in ffs_loc if f.fragment.flow['Name'] == ff.fragment.flow['Name'])
            if ff.fragment.flow['Name'] == ng2.flow['Name']:  # reference fragment: normalization works differently now
                norm = next(x.value for x in ng2.term.term_node.exchange_values(ng2.flow))
            else:
                norm = 1.0
            self.assertAlmostEqual(ff.node_weight, ffl.node_weight / norm, places=12)

    def test_antelope_frag_2_ng_gwp(self):
        if ar is None:
            self.skipTest('antelope archive not available')
        ng = fg['fragments/2']
        res = ng.fragment_lcia(self.gwp).aggregate()
        res_remote = _get_antelope_result(2, 2)
        for key in ('PE', 'US LCI', 'Natural Gas Supply'):
            self.assertAlmostEqual(res[key].cumulative_result, res_remote[key].cumulative_result, places=12)
        self.assertAlmostEqual(res['EI'].cumulative_result, res_remote['EI'].cumulative_result, places=4)
        self.assertNotAlmostEqual(res['EI'].cumulative_result, res_remote['EI'].cumulative_result, places=6)

    def _validate_antelope(self, frag_id):
        frag_l = fg['fragments/%d' % frag_id]
        for lcia_id in (2, 4, 16):
            res_a = _get_antelope_result(frag_id, lcia_id)
            res_l = frag_l.fragment_lcia(self.lcia_map[lcia_id], scenario='quell_eg')
            self.assertAlmostEqual(res_a.total(), res_l.total(), places=10)

    def test_fragment_lcia(self):
        if ar is None:
            self.skipTest('antelope archive not available')
        ar._quiet = False
        for frag_id in (15, 20, 24, 30, 36, 42, 44, 50):
            self._validate_antelope(frag_id)
        ar._quiet = True


if __name__ == '__main__':
    unittest.main()



from collections import defaultdict
from antelope_core.lcia_results import LciaResult, DuplicateResult


class LciaResults(dict):
    """
    A dict of LciaResult objects, with some useful attachments.  The dict gets added to in the normal way, but
    also keeps track of the keys added in sequence, so that they can be retrieved by numerical index.

    The LciaResults object keys should be quantity links

    only keeping this around because I use it in terminations
    """
    def __init__(self, entity, *args, **kwargs):
        super(LciaResults, self).__init__(*args, **kwargs)
        self.entity = entity
        self._scale = 1.0
        self._indices = []

    def __getitem__(self, item):
        """
        __getitem__ can either be used as a numerical index
        :param item:
        :return:
        """
        try:
            return super(LciaResults, self).__getitem__(item)
        except KeyError:
            try:
                int(item)
                return super(LciaResults, self).__getitem__(self._indices[item])
            except (ValueError, TypeError):
                try:
                    return super(LciaResults, self).__getitem__(next(k for k in self.keys() if k.startswith(item)))
                except StopIteration:
                    return LciaResult(None)  #

    def __setitem__(self, key, value):
        assert isinstance(value, LciaResult)
        value.scale_result(self._scale)
        super(LciaResults, self).__setitem__(key, value)
        if key not in self._indices:
            self._indices.append(key)

    '''
    def add(self, value):
        # TODO: add should cumulate components if LciaResult is already present
        assert isinstance(value, LciaResult)
        self.__setitem__(value.quantity.link, value)
    '''

    def indices(self):
        for i in self._indices:
            yield i

    def to_list(self):
        return [self.__getitem__(k) for k in self._indices]

    def scale(self, factor):
        if factor == self._scale:
            return
        self._scale = factor
        for k in self._indices:
            self[k].scale_result(factor)

    def show(self):
        print('LCIA Results\n%s\n%s' % (self.entity, '-' * 60))
        if self._scale != 1.0:
            print('%60s: %10.4g' % ('scale', self._scale))
        for i, q in enumerate(self._indices):
            r = self[q]
            r.scale_result(self._scale)
            print('[%2d] %.3s  %10.5g %s' % (i, q, r.total(), r.quantity))

    def apply_weighting(self, weights, quantity, **kwargs):
        """
        Create a new LciaResult object containing the weighted sum of entries in the current object.

        We want the resulting LciaResult to still be aggregatable. In order to accomplish this, we need to maintain
        all the individual _LciaScores entities in the weighting inputs, and compute their weighted scores here. Then
        we need to log the weighted scores as the *node weights* and use *unit* values of unit scores, because
        SummaryLciaResults are only allowed to be further aggregated if they have the same unit score.

        This feels a bit hacky and may turn out to be a terrible idea.  But there is a certain harmony in making the
        quantity's unit THE unit for a weighting computation. So I think it will work for now.

        :param weights: a dict mapping quantity UUIDs to numerical weights
        :param quantity: EITHER an LcQuantity OR a string to use as the name of in LcQuantity.new()
        :param kwargs: passed to LciaResult
        :return:
        """
        weighted_result = LciaResult(quantity, **kwargs)

        component_list = dict()  # dict maps keys entities
        component_score = defaultdict(float)  # maps keys to weighted scores
        for method, weight in weights.items():
            try:
                result = self.__getitem__(method)  # an LciaResult
            except KeyError:
                continue
            for comp in result.keys():
                if comp in component_list.keys():
                    if result[comp].entity != component_list[comp]:
                        raise DuplicateResult('Key %s matches different entities:\n%s\n%s' % (comp,
                                                                                              result[comp],
                                                                                              component_list[comp]))
                else:
                    component_list[comp] = result[comp].entity
                component_score[comp] += (weight * result[comp].cumulative_result)

        for comp, ent in component_list.items():
            weighted_result.add_component(comp, entity=ent)
            weighted_result.add_summary(comp, ent, component_score[comp], 1.0)

        return weighted_result

    def clear(self):
        super(LciaResults, self).clear()
        self._indices = []

    def update(self, *args, **kwargs):
        super(LciaResults, self).update(*args, **kwargs)
        self._indices = list(self.keys())


class LciaWeighting(object):
    def __init__(self, quantity, weighting):
        """

        :param quantity: a new LcQuantity to represent the weighting
        :param weighting: a weighting dict as defined in apply_weighting
        """
        self._q = quantity
        self._w = weighting

    def weigh(self, res, **kwargs):
        return res.apply_weighting(self._w, self._q, **kwargs)

    def q(self):
        return self._q.link



"""
There are a WHOLE LOT of test cases for fragment traversal because it's a very complex and subtle process.
This is a weakness, ultimately. But the test cases produced here should replace subtlety with clarity.

Here are the things the fragment traversal is supposed to accomplish / enable:

 * sequentially compute node weights, to be multiplied by unit LCIA scores for impact assessment (basic traversal)

 * observability of fragments: each fragment has a separate notion of the exchange value it was "born with" and the
   "observed" exchange value set by the modeler.

   Note: This is a bit of a "solution in search of a problem," but the notional problem is distinguishing between
   inventory data grabbed directly from a reference db and a data point that has been validated by the modeler.
   Traversal with observed=True will use the observed exchange values, which default to 0 so unobserved exchanges
   will be absent.  There is also a routine way to 'observe' a fragment, automatically accepting cached values and
   even recursing.

 * variability of both exchange values and fragment terminations by scenario specification
   (a) exchange value scenario
   (b) termination scenario
   (c) determine valid scenarios for a fragment and subfragments

 * Conservation of up to one quantity per fragment by computing the magnitude of a balance flow

 * the ability to nest fragments and
   (a) have the outcomes of nested traversals affect higher-level traversals (i.e. flow mapping between subfragment
       cutoff flows and superfragment child flows)
   (b) have subfragments driven from non-reference outputs (i.e. encapsulated traversal of sub-fragments)

 * the ability to control whether a subfragment is transparent or opaque (aggregated), a.k.a. 'descend' vs 'non-descend'

 * unit-matching flow magnitude conversions
   (a) when specifying a fragment
   (b) during traversal, between a fragment and its termination

 * the ability to drive fragments in reverse
   (a) by supplying a negative node weight
   (b) by invoking them with a flow whose direction is opposing the fragment's reference direction

Not yet implemented:
 = FragmentFlows must specify which scenarios applied to both ev and termination (exactly 1 for each, maybe None)
   - challenge here is in propagating that info in GhostFragments and aggregations
 = apply-scenario option for subfragment terminations. Build one electricity grid and parameterize it for multiple
   locales; then traverse each instance differently as specified in the termination
 = stochastic exchange values
 = private subfragments can force nondescend?
  + privacy is enforced by providing a fragment ref, which can elect/neglect to implement traverse
"""

import unittest
# from math import floor

from ..fragment_editor import create_fragment
from ...terminations import MissingFlow, FlowConversionError
from antelope_core.entities import LcFlow
from antelope_core.archives import Qdb
from antelope_core.contexts import NullContext
from antelope import CONTEXT_STATUS_

origin = 'test.origin'

qdb = Qdb.new()
qi = qdb.make_interface('quantity')


def new_flow(name, ref_quantity, context=None, **kwargs):
    """

    :param name:
    :param ref_quantity:
    :param context: [None] pending context refactor
    :param kwargs:
    :return:
    """
    if CONTEXT_STATUS_ == 'compat':
        if context is not None and 'compartment' not in kwargs:
            kwargs['compartment'] = str(context)
    ref_q = qi.get_canonical(ref_quantity)
    f = LcFlow.new(name, ref_qty=ref_q, origin=origin, **kwargs)
    # self._archive.add_entity_and_children(f)
    return f


def new_fragment(*args, **kwargs):
    """

    :param args: flow, direction (w.r.t. parent)
    :param kwargs: uuid=None, parent=None, comment=None, value=None, units=None, balance=False;
      **kwargs passed to LcFragment
    :return:
    """
    frag = create_fragment(*args, origin=origin, **kwargs)
    # self._archive.add_entity_and_children(frag)
    return frag


f4_mj_kg = 35
f7_mass = 0.0123

f1 = new_flow('My first flow', 'mass')

mass = f1.reference_entity

# flows

f2 = new_flow('My second flow', 'volume')
f3 = new_flow('My third flow', 'net calorific value')
f3w = new_flow('A waste energy flow', 'net calorific value')
f4 = new_flow('Another mass flow', 'mass')
qi.characterize(f4.link, f4.reference_entity, 'net calorific value', value=f4_mj_kg)
f5 = new_flow('yet another mass flow', 'mass')
f6 = new_flow('An energetic conservation flow', 'net calorific value')
f7 = new_flow('An ancillary flow', 'number of items')
qi.characterize(f7.link, f7.reference_entity, mass, value=f7_mass)
f8 = new_flow('A freight flow', 'freight')
fp = new_flow('A private flow', 'price')

# parameters

# a1
a1_vol = 10
a1_mj_in = 19
a1_addl = 0.88
a1_addl_obs = 0.76
# a1 scenarios
a1_mj_optimistic = a1_mj_in * 0.75
a1_surplus_addl = 1.11

# a2
a2_kwh = 10
a2_mj = a2_kwh * 3.6  # kWh converted to MJ
a2_waste_heat = 5
a2_private = 61
a2_item = 0.003
# a2 scenarios
a2_eff_waste_heat = 3.5
a2_eff_private = 49

# a2_alt
a2_alt_fuel = 1.1 / f4_mj_kg
a2_alt_tx = 2.6

#aa
aa_mj_in = 4
aa_in = 0.14


class FragmentTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """
        Build the test fragments
        :return:
        """
        cls.a1 = new_fragment(f1, 'Output', Name='A Production Process', observe=True)
        cls.a2 = new_fragment(f3, 'Output', value=a2_kwh, units='kWh', Name='A conserving energy conversion process', observe=True)
        cls.af = new_fragment(f4, 'Output', Name='A fuel supply process', observe=True)
        cls.aa = new_fragment(f7, 'Output', Name='An auto-consumption process', observe=True)

        afg = new_fragment(f4, 'Input', parent=cls.af, balance=True, Name='Internal fuel supply')
        afg.terminate(NullContext)  # this is the replacement for to_foreground()

        new_fragment(f2, 'Input', parent=cls.a1, value=a1_vol)  # not observed
        a3 = new_fragment(f3, 'Input', parent=cls.a1, value=a1_mj_in, observe=True)
        a3.set_exchange_value('optimistic', a1_mj_optimistic)
        a3.terminate(cls.a2, descend=False)

        new_fragment(f7, 'Input', parent=a3)
        new_fragment(f5, 'Input', parent=cls.a1, balance=True).terminate(cls.af, term_flow=f4)  # to be run negated
        aq = new_fragment(f4, 'Input', parent=cls.a1, value=a1_addl)
        aq.observe(value=a1_addl_obs)
        aq.observe(a1_surplus_addl, scenario='surplus')

        '''a1
           -<--O   70a01 [       1 kg] A Production Process
            [   1 unit] A Production Process
               | -<--#:: 3fe31 (=      1 kg) yet another mass flow
               | -<----: a37a3 (     1.2 kg) Another mass flow
               | -<----: a4a0c (      10 m3) My second flow
               | -<--#   397f5 (      19 MJ) My third flow
               |  [   1 unit] My third flow
               |     | -<----: c3f26 (       1 Item(s)) An ancillary flow
               |     x 
               x 
        (a1d is the same but with descend=True on "My third flow")
        '''

        a3w = new_fragment(f3w, 'Output', parent=cls.a2, value=a2_waste_heat, observe=True)
        a3w.set_exchange_value('efficiency', a2_eff_waste_heat)
        a3w.terminate(NullContext)  # this is the replacement for to_foreground()

        cls.a6 = new_fragment(f6, 'Input', parent=cls.a2, balance=True)
        cls.a6.terminate(cls.af, term_flow=f4)
        a2p = new_fragment(fp, 'Input', parent=cls.a2, value=a2_private)
        a2p.set_exchange_value('efficiency', a2_eff_private)
        new_fragment(f7, 'Input', parent=a2p, value=a2_item)

        cls.a2.observe(accept_all=True, recurse=True)
        '''a2
           -<--O   a33fc [      36 MJ] A conserving energy conversion process
            [   1 unit] A conserving energy conversion process
               | -<--O   53279 [      61 EUR] A private flow
               |  [   1 unit] A private flow
               |     | -<----: be6fa [   0.003 Item(s)] An ancillary flow
               |     x 
               | =>=-O   b1951 [       5 MJ] A waste energy flow
               | -<--#:: b09e3 (=      1 MJ) An energetic conservation flow
               x 
        '''
        # make an alternate energy supply fragment to test scenario terminations
        cls.a2_alt = new_fragment(f3, 'Output', Name='An alternate energy conversion process', observe=True)
        new_fragment(f4, 'Input', parent=cls.a2_alt, value=a2_alt_fuel, observe=True)
        new_fragment(f8, 'Input', parent=cls.a2_alt, value=a2_alt_tx, observe=True)
        '''a2_alt
           -<--O   db333 [       1 MJ] An alternate energy conversion process
            [   1 unit] An alternate energy conversion process
               | -<----: 99b68 (  0.0314 kg) Another mass flow
               | -<----: beda8 (     2.6 t*km) A freight flow
               x 
        This a2_alt is supplied as an alternate termination scenario. it lacks the 'ancillary item' flow so that
        child flow should be found to be zero and should not appear in the traversal.  At the same time, it contains
        a 'freight' flow which SHOULD be visible as long as descend=True.
        '''

        a3.terminate(cls.a2_alt, scenario='improvement', descend=True)

        new_fragment(f5, 'Input', parent=cls.aa, value=aa_in, observe=True)
        ac = new_fragment(f3, 'Input', value=aa_mj_in, parent=cls.aa, observe=True)
        new_fragment(f4, 'Output', parent=cls.aa, balance=True).terminate(cls.af)  # to be run in reverse
        ac.terminate(cls.a2)
        ac.terminate(cls.a2, scenario='nondescend', descend=False)  # to be run in reverse
        '''aa
           -<--O   b0565 [       1 Item(s)] An auto-consumption process
            [   1 unit] An auto-consumption process
               | -<----: dbda7 (    0.14 kg) yet another mass flow
               | -<--#:: 332f7 (       4 MJ) My third flow
               | =>=-#:: 48149 (=      1 kg) Another mass flow
               x 
               
        This process produces ancillary items, but uses fragment a2 which consumes ancillary items. This process
        also produces a fuel co-product which is used to drive the fuel subfragment negative.
        '''

    def _check_fragmentflows(self, ffs, flow, direction, *magnitudes):
        """
        A routine to test whether a set of traversal results includes the expected set of flows, directions, and
        magnitudes.

        With our change in which we no longer include subfragments in traversal, the easiest way to maintain integrity
        of the test is to replicate the old behavior, which means extracting in-line subfragments, but scaling magnitude

        :param ffs: a list of fragment flows
        :param flow: a flow entity to look for
        :param direction: the direction to look for
        :param magnitudes: a sequence of positional args indicating the magnitudes expected to be found in the traversal
        :return:
        """
        filt_list = []
        filt_mag = []
        for ff in ffs:
            if ff.fragment.flow == flow and ff.fragment.direction == direction:
                filt_list.append(ff)
                filt_mag.append(ff.magnitude)
            if ff.term.descend:
                for k in ff.subfragments:
                    if k.fragment.flow == flow and k.fragment.direction == direction:
                        filt_list.append(k)
                        filt_mag.append(k.magnitude * ff.node_weight)
        self.assertEqual(len(filt_list), len(magnitudes))
        self.assertSetEqual(set(filt_mag), set(magnitudes))

    def test_af_traversal(self):
        """
        af is a very simple fragment with one reference flow and one foreground flow, which is a stand-in for an
        aggregated process inventory.
        :return:
        """
        ffs = self.af.traverse(None)
        self._check_fragmentflows(ffs, f4, 'Input', 1, 1)

    def test_af_unit_flows(self):
        """
        unit flows reports two separate sets of fragment flows: cutoffs (cross system boundary) and nodes (within
        system boundary).  For af, there is one cutoff (the reference flow) and two internal, both foreground fragments.
        :return:
        """
        io, ff = self.af.unit_flows()
        self._check_fragmentflows(io, f4, 'Output', 1)
        self.assertEqual(len(ff), 2)
        # foreground <=> term is self -- to_foreground() replaced with termination to NullContext
        self.assertSetEqual({self.af, NullContext}, {f.term.term_node for f in ff})

    def test_unit_conversion_on_fragment_creation(self):
        self.assertEqual(self.a2.exchange_value(), a2_mj)

    def test_unit_conversion_on_flow_termination(self):
        """
        This is testing the traversal across the termination of a6, which should perform a ref quantity conversion
        from MJ (a6.flow.unit()) to kg (af.flow.unit()) and have a value of [the balance net] divided by [f4_mj_kg]
        :return:
        """
        self.assertEqual(self.a6.term.flow_conversion, 1.0 / f4_mj_kg)
        proper_kg = (a2_mj + a2_waste_heat) / f4_mj_kg
        ff = self.a2.traverse(None)
        # we can no longer filter on traversal because subfragments are hidden
        ff_a6 = next(f for f in ff if f.fragment is self.a6)
        self.assertEqual(ff_a6.node_weight, proper_kg)

    def test_subfragment_child(self):
        """
        Child flows from a subfragment termination have their e.v.s computed based on the subfragment traversal
        self.a1 subfragment of self.a2 with child flow f7 'An ancillary flow'
        :return:
        """
        a2_ancillary = a2_item * a2_private
        a1_ancillary = a1_mj_in / self.a2.exchange_value() * a2_ancillary
        self._check_fragmentflows(self.a2.traverse(None), f7, 'Input', a2_ancillary)
        self._check_fragmentflows(self.a1.traverse(None), f7, 'Input', a1_ancillary)

    def test_missing_flow(self):
        k = create_fragment(f6, 'Output')
        with self.assertRaises(MissingFlow):
            k.terminate(self.a1, term_flow=f6)
        with self.assertRaises(FlowConversionError):
            k.terminate(self.a1)  # we now permit the flow to be terminated without an operable flow conversion,
            k.traverse()  # but it cannot be traversed

    def test_unobserved_traversal(self):
        """
        A simple unobserved traversal should just result in cached evs for observable flows, properly computed balance
        Note: these tests rely on the behavior that LcFragment.traverse() by default uses cached (not observed) EVs
        This contrasts with traversal via the foreground interface, which uses observed EVs by default

        :return:
        """
        ffuobs = self.a1.traverse()
        bal = self.a1.exchange_value() - a1_addl
        for c in self.a1.child_flows:
            ff = next(f for f in ffuobs if f.fragment is c)
            if c.is_balance:
                self.assertEqual(ff.magnitude, bal)
                # self._check_fragmentflows(ffuobs, c.flow, c.direction, bal)
            else:
                self.assertEqual(ff.magnitude, c.cached_ev)
                # self._check_fragmentflows(ffuobs, c.flow, c.direction, c.cached_ev)

    def test_unobservable_balance(self):
        for f in self.a1.child_flows:
            if f.is_balance:
                self.assertFalse(f.observable())
            else:
                self.assertTrue(f.observable())

    def test_observed_traversal(self):
        """
        An unobserved traversal should use the cached values (i.e. defined at creation) whereas an observed traversal
        should use the modeler's observed values, which default to zero.  Fragment a1 is not observed whereas fragment
        a2 is auto-observed to apply cached values as observed values throughout.
        Note: these tests rely on the behavior that LcFragment.traverse() by default uses cached (not observed) EVs
        This contrasts with traversal via the foreground interface, which uses observed EVs by default

        :return:
        """
        ff1 = self.a1.cutoffs()
        ff1o = self.a1.cutoffs(observed=True)
        ff2 = self.a2.cutoffs()
        ff2o = self.a2.cutoffs(observed=True)
        self.assertEqual(ff2, ff2o)
        self.assertNotEqual(ff1, ff1o)
        ff1t = self.a1.traverse(observed=True)
        fbal = next(f for f in ff1t if f.fragment is self.a1.balance_flow)
        self.assertEqual(fbal.magnitude, self.a1.exchange_value() - a1_addl_obs)

    def test_scenarios_detection(self):
        self.assertSetEqual({k for k in self.a1.scenarios()}, {'surplus', 'optimistic', 'improvement', 'efficiency'})
        self.assertSetEqual({k for k in self.a2.scenarios()}, {'efficiency'})
        self.assertSetEqual({k for k in self.aa.scenarios()}, {'efficiency', 'nondescend'})
        self.assertSetEqual({k for k in self.af.scenarios()}, set())

    def test_scenario_ev(self):
        """
        surplus
        :return:
        """
        default = self.a1.traverse()
        surplus = self.a1.traverse('surplus')
        self._check_fragmentflows(default, f4, 'Input', a1_addl, 1-a1_addl, 1-a1_addl)
        self._check_fragmentflows(surplus, f4, 'Input', a1_surplus_addl, 1 - a1_surplus_addl, 1 - a1_surplus_addl)

    def test_scenario_termination(self):
        default = self.a1.traverse()
        improved = self.a1.traverse('improvement')
        self.assertIn(self.a2_alt, [f.term.term_node for f in improved])
        self.assertNotIn(self.a2_alt, [f.term.term_node for f in default])
        self._check_fragmentflows(default, f8, 'Input')  # not present
        self._check_fragmentflows(improved, f8, 'Input', a1_mj_in * a2_alt_tx)  # present

    def test_autoconsumption(self):
        """
        fragment self.aa produces f7 "ancillary flows" in the amount of one unit.  but it also depends on fragment
        self.a2, which has a cutoff input requirement for f7.  The inventory results for self.aa should report only
        net production of f7
        :return:
        """
        f7_autoconsumption = aa_mj_in / self.a2.exchange_value() * a2_private * a2_item
        inv = self.aa.cutoffs()
        f7_out = next(f for f in inv if f.flow is f7)
        f5_in = next(f for f in inv if f.flow is f5)
        self.assertEqual((f7_out.direction, f7_out.value), ('Output', 1 - f7_autoconsumption))
        self.assertEqual((f5_in.direction, f5_in.value), ('Input', aa_in))

    def test_nonreference_subfragment(self):
        """
        A termination to a non-reference flow of a subfragment computes the proper node weight for the subfragment
        Here we build an ad hoc fragment that coproduces f5 'yet another mass flow'-- which is also
        required by the autoconsumption fragment self.aa.

        We are going to assume that the co-product drives additional production of f7 "ancillary flows" by terminating
        our coproduct flow to self.aa.  This should have the effect of driving the autoconsuming fragment in its
        forward direction and outputting f7.
        :return:
        """
        coproduct = 6
        fq = new_flow('transparent aluminum', 'area')
        at = new_fragment(fq, 'Output')
        new_fragment(f5, 'Output', parent=at, value=coproduct).terminate(self.aa, term_flow=f5)
        # node weight of driven node
        aa_nw = coproduct / aa_in
        # unit consumption of f7 by energy production node
        f7_autoconsumption = aa_mj_in / self.a2.exchange_value() * a2_private * a2_item
        net_f7_production = aa_nw * (1 - f7_autoconsumption)
        ffs = at.traverse()
        self._check_fragmentflows(ffs, f7, 'Output', net_f7_production)

    def test_negative_fragment(self):
        """
        self.a1 balancing flow is negative under the 'surplus' scenario.  self.a1 balancing flow terminates to self.af
        so if the driving amount is negative, the node weight of the subfragment should be negative.
        :return:
        """
        ffs = self.a1.traverse('surplus')
        self._check_fragmentflows(ffs, f5, 'Input', 1 - a1_surplus_addl)
        for f in ffs:
            if f.fragment.top() is self.af:
                self.assertTrue(f.node_weight < 0)

    def test_balance_nonreference_quantities(self):
        """
        self.aa reference product is characterized w.r.t. mass, so balance flow should account for that
        :return:
        """
        ffs = self.aa.traverse()
        net_balance = aa_in - mass.cf(f7)
        self._check_fragmentflows(ffs, f4, 'Output', net_balance)

    def test_inverted_fragment(self):
        """
        self.aa produces an output of f5 which is terminated to self.af, which also produces an output of af.  Thus
        when self.af is the subfragment its node weight should be negative.
        :return:
        """
        ffs = self.aa.traverse('nondescend')  # ensure that the subfragment's dependence on af is concealed
        for f in ffs:
            if f.fragment.top() is self.af:
                self.assertTrue(f.node_weight < 0, '%s' % f)

    def test_descend_equivalence(self):
        """
        Cutoffs results should not vary regardless of descend (this may require more robust testing)
        :return:
        """
        inv_d = self.aa.cutoffs()
        inv_nd = self.aa.cutoffs('nondescend')
        self.assertEqual(inv_d, inv_nd)

    def test_nondescend_privacy(self):
        """
        The internal fragment of fp in self.a2 should be concealed when descend=False, visible when descend=True
        :return:
        """
        ff_d = self.aa.traverse()
        ff_nd = self.aa.traverse('nondescend')
        expected_private = aa_mj_in / self.a2.exchange_value() * a2_private
        expected_waste = aa_mj_in / self.a2.exchange_value() * a2_waste_heat
        self._check_fragmentflows(ff_d, fp, 'Input', expected_private)
        self._check_fragmentflows(ff_nd, fp, 'Input')
        self._check_fragmentflows(ff_d, f3w, 'Output', expected_waste)
        self._check_fragmentflows(ff_nd, f3w, 'Output')

    def test_nonscenario(self):
        """
        A Scenarios specification is just an unordered collection of scenario names. Names for scenarios that are not
        found in the fragment should simply have no effect.
        :return:
        """
        self.assertNotIn('moriarty', [k for k in self.a1.scenarios()])
        ffs = self.a1.traverse(observed=True)
        ffs_m = self.a1.traverse('moriarty')
        self.assertEqual(ffs, ffs_m)

    def test_multiple_scenarios(self):
        """
        Here we combine several different scenarios and look for signatures of each in the traversal results.
        'optimistic' reduces a1 energy requirements
        'efficiency' improves a2 performance
        'surplus' changes the amount of input received by a1
        'improvement' substitutes a2_alt for a2
        :return:
        """
        ff_o_e = self.a1.traverse({'optimistic', 'efficiency'})
        expected_item = (a1_mj_optimistic / self.a2.exchange_value()) * (a2_eff_private * a2_item)
        self._check_fragmentflows(ff_o_e, f7, 'Input', expected_item)
        self._check_fragmentflows(ff_o_e, f5, 'Input', 1-a1_addl_obs)

        ff_o_s_e = self.a1.traverse({'optimistic', 'surplus', 'efficiency'})
        self._check_fragmentflows(ff_o_s_e, f7, 'Input', expected_item)
        self._check_fragmentflows(ff_o_s_e, f5, 'Input', 1 - a1_surplus_addl)

        ff_o_i_s = self.a1.traverse({'optimistic', 'improvement', 'surplus'})
        expected_another = a1_mj_optimistic / self.a2_alt.exchange_value() * a2_alt_fuel
        self._check_fragmentflows(ff_o_i_s, f7, 'Input')
        self._check_fragmentflows(ff_o_i_s, f4, 'Input',
                                  a1_surplus_addl, 1 - a1_surplus_addl, 1 - a1_surplus_addl, expected_another)

    def test_tuple_scenarios(self):
        expected_item = a1_mj_optimistic / self.a2.exchange_value() * (a2_eff_private * a2_item)
        ff_o_s_e = self.a1.traverse(('optimistic', 'surplus', 'efficiency'))
        self._check_fragmentflows(ff_o_s_e, f7, 'Input', expected_item)
        self._check_fragmentflows(ff_o_s_e, f5, 'Input', 1 - a1_surplus_addl)




if __name__ == '__main__':
    unittest.main()
